
function AbstractPlotting.convert_attribute(x, ::Nothing, key1, key2)
    return convert_attribute(x, key1, key2)
end

function AbstractPlotting.convert_attribute(x::AbstractVector, ::Nothing, ::key"linewidth",
                                            key2)
    return x[1:2:(length(x) - 1)]
end

topoint(x::AbstractVector{Point{N,Float32}}) where {N} = x

# GRRR STUPID SubArray, with eltype different from getindex(x, 1)
topoint(x::SubArray) = topoint([el for el in x])

function topoint(x::AbstractArray{<:Point{N,T}}) where {T,N}
    return topoint(Point{N,Float32}.(x))
end
function topoint(x::AbstractArray{<:Tuple{P,P}}) where {P<:Point}
    return topoint(reinterpret(P, x))
end

function create_shader(scene::Scene, plot::Union{Lines,LineSegments})
    # Potentially per instance attributes
    positions = lift(plot[1]) do points
        points = topoint(points)
        if plot isa LineSegments
            return points
        else
            # Repeat every second point to connect the lines !
            return topoint(TupleView{2, 1}(points))
        end
    end
    startr = lift(p -> 1:2:(length(p) - 1), positions)
    endr = lift(p -> 2:2:length(p), positions)
    p_start_end = lift(positions) do positions
        return (positions[startr[]], positions[endr[]])
    end

    per_instance = Dict{Symbol,Any}(:segment_start => Buffer(lift(first, p_start_end)),
                                    :segment_end => Buffer(lift(last, p_start_end)))
    uniforms = Dict{Symbol,Any}()
    for k in (:linewidth, :color)
        attribute = lift(plot[k]) do x
            x = convert_attribute(x, Key{k}(), key"lines"())
            if plot isa LineSegments
                return x
            else
                # Repeat every second point to connect the lines!
                return isscalar(x) ? x : reinterpret(eltype(x), TupleView{2, 1}(x))
            end
        end
        if isscalar(attribute)
            uniforms[k] = attribute
            uniforms[Symbol("$(k)_start")] = attribute
            uniforms[Symbol("$(k)_end")] = attribute
        else
            if attribute[] isa AbstractVector{<:Number} && haskey(plot, :colorrange)
                attribute = lift(array2color, attribute, plot.colormap, plot.colorrange)
            end
            per_instance[Symbol("$(k)_start")] = Buffer(lift(x -> x[startr[]], attribute))
            per_instance[Symbol("$(k)_end")] = Buffer(lift(x -> x[endr[]], attribute))
        end
    end

    uniforms[:resolution] = scene.camera.resolution
    uniforms[:model] = plot.model
    positions = meta(Point2f0[(0, -1), (0, 1), (1, -1), (1, 1)],
                     uv=Vec2f0[(0, 0), (0, 0), (0, 0), (0, 0)])
    instance = GeometryBasics.Mesh(positions, GLTriangleFace[(1, 2, 3), (2, 4, 3)])
    return InstancedProgram(WebGL(), lasset("line_segments.vert"),
                            lasset("line_segments.frag"), instance,
                            VertexArray(; per_instance...); uniforms...)
end

function draw_js(jsctx, jsscene, mscene::Scene, plot::Lines)
    @extract plot (color, linewidth, model, transformation)
    positions = plot[1]
    colors_converted = lift(x -> to_color(x), color)
    mesh = jslines!(jsctx, jsscene, plot, positions, colors_converted, linewidth, model)
    map(plot.visible) do visible
        return mesh.visible = visible
    end
    return mesh
end

br_view(scalar, idx) = scalar
br_view(array::AbstractVector, idx) = view(array, idx)

"""
    non_nan_segments_0(vector::AbstractVector)

Returns 0 based indices for all line segments that aren't NaN
"""
non_nan_segments_0(vector::AbstractVector) = non_nan_segments_0(vector, UInt32[])
function non_nan_segments_0!(vector::AbstractVector, indices_0::Vector{UInt32})
    empty!(indices_0)
    @inbounds for i in 1:(length(vector) - 1)
        if !(isnan(vector[i]) || isnan(vector[i + 1]))
            push!(indices_0, UInt32(i - 1), UInt32(i))
        end
    end
    return indices_0
end

function jslines!(THREE, scene, plot, positions_nan, colors, linewidth, model, typ=:lines)
    geometry = THREE.new.BufferGeometry()
    opaque_color = "#000"
    opacity = 1
    color_buffer = nothing
    positions_buff = Float32[]
    positions_flat = map(positions_nan) do positions
        f32 = reinterpret(Float32, positions)
        resize!(positions_buff, length(f32))
        positions_buff[:] .= f32
        return positions_buff
    end
    index_buffer = UInt32[]
    segments_ui32_0 = map(positions_nan) do positions
        return non_nan_segments_0!(positions, index_buffer)
    end

    if colors[] isa Colorant
        opaque_color = "#" * hex(color(colors[]))
        opacity = alpha(colors[])
    else
        flat = flatten_buffer(colors[])
        color_buffer = THREE.new.Float32BufferAttribute(flat, 4)
        geometry.setAttribute("color", color_buffer)
    end
    material = THREE.new.LineBasicMaterial(; color=opaque_color, opacity=opacity,
                                           transparent=true)

    nd = length(eltype(positions_nan[]))
    position_buffer = THREE.new.Float32BufferAttribute(positions_flat[], nd)
    geometry.setAttribute("position", position_buffer)
    geometry.setIndex(segments_ui32_0[])

    onjs(THREE, positions_flat, js"""function (flat_positions){
        var flat = deserialize_js(flat_positions);
        var position_buffer = $(position_buffer);
        position_buffer.set(flat, 0);
        position_buffer.needsUpdate = true;
    }""")

    onjs(THREE, segments_ui32_0, js"""function (segments){
        var indices = deserialize_js(segments);
        var geometry = $(geometry);
        geometry.setIndex(indices);
    }""")

    # ThreeJS does culling by calculating the boundingsphere
    # But somehow it doesn't play well with our custom attributes.
    # We just set it to a huge sphere, to not get them culled.
    # Would be nice to just set them to null or so, but that's
    # what triggers threejs to actually calculate the sphere
    geometry.boundingSphere = THREE.new.Sphere()
    geometry.boundingSphere.radius = 10000000000000f0

    mesh = THREE.new.LineSegments(geometry, material)
    mesh.matrixAutoUpdate = false
    mesh.matrix.set(model[]...)
    scene.add(mesh)
    update_model!(mesh, plot)
    return mesh
end
